package com.me.brickbuster.state;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.glutils.ShapeRenderer.ShapeType;
import com.badlogic.gdx.math.MathUtils;
import com.badlogic.gdx.math.Vector2;
import com.badlogic.gdx.physics.box2d.*;
import com.me.brickbuster.BrickBuster;
import com.me.brickbuster.entity.Ball;
import com.me.brickbuster.entity.Brick;
import com.me.brickbuster.entity.Paddle;
import com.me.brickbuster.entity.powerup.*;
import com.me.brickbuster.physics.Box2dContactListener;

import java.util.*;

public class PlayState extends State {

	public static final float PIXEL_PER_METER = 40f;    // Board dimension: 54x96 meters
	public static final float BOARD_WIDTH = 54f;
	public static final float BOARD_HEIGHT = 96f;
	public static final float EDGE_PADDING = .125f;


	public static final float SHIELD_HEIGHT = 0.75f;
	public static final float POWERUP_CHANCE = 0.15f;

	public static final int COLUMNS = 9;
	public static final int ROWS = 8;

	public static final Vector2 lowerLeftCorner =
			new Vector2(EDGE_PADDING,EDGE_PADDING);
	public static final Vector2 lowerRightCorner =
			new Vector2(BrickBuster.BOARD_WIDTH/PIXEL_PER_METER-EDGE_PADDING,EDGE_PADDING);
	public static final Vector2 upperRightCorner =
			new Vector2(BrickBuster.BOARD_WIDTH/PIXEL_PER_METER-EDGE_PADDING, BrickBuster.BOARD_HEIGHT/PIXEL_PER_METER-EDGE_PADDING);
	public static final Vector2 upperLeftCorner =
			new Vector2(EDGE_PADDING, BrickBuster.BOARD_HEIGHT/PIXEL_PER_METER-EDGE_PADDING);

	public World world;
	public Body playArea;

	public List<PowerUp> powerUps;
	public Paddle paddle;
	public List<Ball> balls;
	public List<Brick> bricks;

	private int shieldCount = 0;
	private float updateTime = 0f;

	private final Box2DDebugRenderer debugRenderer;

	public PlayState(BrickBuster game) {
		super(game);
		debugRenderer = new Box2DDebugRenderer();
	}

	@Override
	public void setup() {
		// Initialize a world with no gravity
		world = new World(new Vector2(), false);
		world.setContactListener(new Box2dContactListener());

		// define a playArea body with position set to 0
		BodyDef playAreaDef = new BodyDef();
		playAreaDef.position.set(new Vector2());
		EdgeShape screenEdge = new EdgeShape();

		playArea = world.createBody(playAreaDef);
		// Right edge
		screenEdge.set(lowerRightCorner, upperRightCorner);
		playArea.createFixture(screenEdge, 0f);
		// Top edge
		screenEdge.set(upperRightCorner, upperLeftCorner);
		playArea.createFixture(screenEdge, 0f);
		// Left edge
		screenEdge.set(upperLeftCorner, lowerLeftCorner);
		playArea.createFixture(screenEdge, 0f);
		// Bottom edge
		screenEdge.set(lowerLeftCorner, lowerRightCorner);
		playArea.createFixture(screenEdge, 0f);
		screenEdge.dispose();

		powerUps = new ArrayList<PowerUp>();
		paddle = new Paddle(this);

		float brick_padding = ((BrickBuster.BOARD_WIDTH/PIXEL_PER_METER) - COLUMNS * Brick.BRICK_WIDTH) / (COLUMNS + 1);
		bricks = new ArrayList<Brick>();
		for (int col = 0; col < COLUMNS; col++) {
			for (int row = 0; row < ROWS; row++) {
				float x = brick_padding + (col * (Brick.BRICK_WIDTH + brick_padding));
				float y = brick_padding + Brick.BRICK_HEIGHT + (row * (Brick.BRICK_HEIGHT + brick_padding));

				PowerUpType powerUpType = null;
				if (MathUtils.randomBoolean(POWERUP_CHANCE)) {
					powerUpType = PowerUpType.getWeightedRandom();
				}

				bricks.add(new Brick(this, powerUpType, x, BrickBuster.BOARD_HEIGHT/PIXEL_PER_METER - y));
			}
		}

		balls = new ArrayList<Ball>();
		balls.add(new Ball(this));
	}

	@Override
	public void render() {
		long start = System.nanoTime();
		for (Brick block : bricks) {
			block.render(game.sr);
		}
		for (PowerUp powerUp : powerUps) {
			powerUp.render(game.sr);
		}
		for (Ball ball : balls) {
			ball.render(game.sr);
		}
		paddle.render(game.sr);

		if (getShieldCount() > 0) {
			game.sr.begin(ShapeType.Filled);
			game.sr.setColor(Color.SALMON);
			game.sr.rect(0, 0, BrickBuster.BOARD_WIDTH, getShieldCount() * SHIELD_HEIGHT);
			game.sr.end();
		}
		long renderTime = System.nanoTime() - start;

		game.sb.begin();
		game.font.setColor(Color.GRAY);
		game.font.draw(game.sb, String.format("FPS: %d Update: %.2f ms Render: %.2f ms",
				Gdx.graphics.getFramesPerSecond(), updateTime/1000000f, renderTime/1000000f), 10, BrickBuster.BOARD_HEIGHT-10);
		game.sb.end();

		debugRenderer.render(world, game.cam.combined.cpy().scl(PIXEL_PER_METER));
	}

	@Override
	public void update(float dt) {
		long start = System.nanoTime();
		world.step(dt, 6, 2);

		paddle.update(dt);

		for (Iterator<Ball> it = balls.iterator(); it.hasNext();) {
			Ball ball = it.next();
			ball.update(dt);
			if (ball.isDead()) {
				it.remove();
			}
		}
		if (balls.isEmpty()) {
			ballReset();
		}

		for (Iterator<PowerUp> it = powerUps.iterator(); it.hasNext();) {
			PowerUp powerUp = it.next();
			powerUp.update(dt);
			if(powerUp.isCaught()) {
				it.remove();
			}
		}

		if (bricks.isEmpty()) {
			game.setScreen(new MenuState(game));
			dispose();
		}
		updateTime = System.nanoTime() - start;
	}

	@Override
	public void dispose() {
		super.dispose();

		powerUps.clear();
		powerUps = null;
		balls.clear();
		balls = null;
		bricks.clear();
		bricks = null;
		paddle = null;
	}

	public int getShieldCount() {
		return shieldCount;
	}

	public void addShield() {
		shieldCount++;
		paddle.setY(paddle.getY() + SHIELD_HEIGHT);
		for (Ball ball : balls) {
			if (ball.isStuck()) {
				ball.setY(ball.getY() + SHIELD_HEIGHT);
			}
		}
	}

	public void removeShield() {
		shieldCount--;
		paddle.setY(paddle.getY() - SHIELD_HEIGHT);
		for (Ball ball : balls) {
			if (ball.isStuck()) {
				ball.setY(ball.getY() - SHIELD_HEIGHT);
			}
		}
	}

	public void ballReset() {
		Ball ball = new Ball(this);
		ball.setX(paddle.getX());
		ball.setY(Paddle.PADDLE_Y + Paddle.PADDLE_HEIGHT + Ball.RADIUS);
		balls.add(ball);
		paddle.setSticky(false);
		paddle.setWidth(Paddle.DEFAULT_WIDTH);
	}

}
